"""Mixin for API objects that can be converted to a dict compatible with the Elastiknn Json API."""

from typing import Dict


class Codec(object):
    """Mixin for API objects that can be converted to a dict compatible with the Elastiknn Json API."""

    def to_dict(self) -> Dict:
        """Return a dictionary compatible with the Elastiknn Json API."""
        raise NotImplementedError




import json
from concurrent.futures import wait
from concurrent.futures.thread import ThreadPoolExecutor
from elasticsearch import Elasticsearch
from logging import Logger
from time import time
from typing import List, Union

import numpy as np
from scipy.sparse import csr_matrix
from tqdm import tqdm

from . import ELASTIKNN_NAME
from .api import Mapping, Vec, NearestNeighborsQuery, Similarity
from .client import ElastiKnnClient
from .utils import canonical_vectors_to_elastiknn, valid_metrics_algos


class ElastiknnModel(object):

    def __init__(self, algorithm='exact', metric='jaccard', es: Elasticsearch = None, mapping_params: dict = None,
                 query_params: dict = None, n_jobs=1, index: str = None):
        self._logger = Logger(self.__class__.__name__)
        self._n_jobs = n_jobs
        self._tpex = ThreadPoolExecutor(self._n_jobs)
        self._field = "vec"
        self._index = index

        assert (algorithm, metric) in valid_metrics_algos, \
            f"algorithm and metric should be one of {valid_metrics_algos}"

        field = "vec"
        dummy = Vec.Indexed("", "", "")

        def _mk_mapping_query(dims: int) -> (Mapping.Base, NearestNeighborsQuery.Base):
            if algorithm == "exact":
                if metric == 'l1':
                    return (Mapping.DenseFloat(dims), NearestNeighborsQuery.Exact(field, dummy, Similarity.L1))
                elif metric == 'l2':
                    return (Mapping.DenseFloat(dims), NearestNeighborsQuery.Exact(field, dummy, Similarity.L2))
                elif metric == 'angular':
                    return (Mapping.DenseFloat(dims), NearestNeighborsQuery.Exact(field, dummy, Similarity.L2))
                elif metric == 'jaccard':
                    return (Mapping.SparseBool(dims), NearestNeighborsQuery.Exact(field, dummy, Similarity.Jaccard))
                elif metric == 'hamming':
                    return (Mapping.SparseBool(dims), NearestNeighborsQuery.Exact(field, dummy, Similarity.Hamming))
            elif algorithm == 'sparse_indexed':
                if metric == 'jaccard':
                    return (Mapping.SparseIndexed(dims), NearestNeighborsQuery.SparseIndexed(field, dummy, Similarity.Jaccard))
                elif metric == 'hamming':
                    return (Mapping.SparseIndexed(dims), NearestNeighborsQuery.SparseIndexed(field, dummy, Similarity.Hamming))
            elif algorithm == 'lsh':
                if metric == 'jaccard':
                    return (Mapping.JaccardLsh(dims, **mapping_params), NearestNeighborsQuery.JaccardLsh(field, dummy, **query_params))
            raise NameError

        self._mk_mapping_query = lambda dims: _mk_mapping_query(dims)
        self._mapping = None
        self._query = None
        self._eknn = ElastiKnnClient(es)

    def fit(self, X: Union[np.ndarray, csr_matrix, List[Vec.SparseBool], List[Vec.DenseFloat]], shards: int = 1):
        vecs = list(canonical_vectors_to_elastiknn(X))
        dims = len(vecs[0])
        (self._mapping, self._query) = self._mk_mapping_query(dims)

        if self._index is None:
            self._index = f"{ELASTIKNN_NAME}-{int(time())}"
            self._logger.warning(f"index was not given, using {self._index} instead")

        self._eknn.es.indices.delete(self._index, ignore=[400, 404])
        body = dict(settings=dict(number_of_shards=shards, number_of_replicas=0))
        self._eknn.es.indices.create(self._index, body=json.dumps(body))
        self._eknn.es.indices.refresh(self._index)
        self._eknn.put_mapping(self._index, self._field, mapping=self._mapping)
        self._eknn.es.indices.refresh(self._index)

        self._logger.info(f"indexing {len(vecs)} vectors into index {self._index}")
        ids = [str(i + 1) for i in range(len(vecs))]  # Add one because 0 is an invalid id in ES.
        self._eknn.index(self._index, self._field, vecs, ids, refresh=True)

    def kneighbors(self, X: Union[np.ndarray, csr_matrix, List[Vec.SparseBool], List[Vec.DenseFloat], List[Vec.Base]],
                   n_neighbors: int, return_similarity: bool = False, allow_missing: bool = False, progbar: bool = False):
        mapped = self._tpex.map(
            lambda v: self._eknn.nearest_neighbors(self._index, self._query.with_vec(v), n_neighbors, False),
            canonical_vectors_to_elastiknn(X))
        inds = np.ones((len(X), n_neighbors), dtype=np.int32) * -1
        sims = np.zeros((len(X), n_neighbors), dtype=np.float) * np.nan
        for i, res in tqdm(enumerate(mapped), disable=not progbar):
            hits = res['hits']['hits']
            assert allow_missing or len(hits) == n_neighbors, f"Expected {n_neighbors} hits for vector {i} but got {len(hits)}"
            for j, hit in enumerate(hits):
                inds[i][j] = int(hit['_id']) - 1  # Subtract one because 0 is an invalid id in ES.
                sims[i][j] = float(hit['_score'])

        if return_similarity:
            return inds, sims
        else:
            return inds



"""Client for using Elastiknn."""

from typing import Iterable, Tuple, Dict

from elasticsearch import Elasticsearch
from elasticsearch.helpers import bulk

from .api import *


class ElastiKnnClient(object):

    def __init__(self, es: Elasticsearch = None):
        """Wrapper on the official `elasticsearch.Elasticsearch` client for making Elastiknn requests.

        The client is fairly primitive in that it assumes the vector is basically the only thing stored in each doc.
        For more complex use-cases you should use this client as an example for constructing the mapping, indexing,
        and search requests in your own code.

        Parameters
        ----------
        es : `elasticsearch.Elasticsearch` client.
            This client is used internally to make all requests.
            Defaults to a client pointing at http://localhost:9200.
        """
        if es is None:
            self.es = Elasticsearch(["http://localhost:9200"])
        else:
            self.es = es

    def put_mapping(self, index: str, field: str, mapping: Mapping.Base):
        """
        Update the mapping at the given index and field to store an Elastiknn vector.

        Parameters
        ----------
        index : string
            Index containing the given field. Should already exist.
        field : string
            Field that should use the given mapping.
        mapping : instance of `Mapping.Base`
            Mapping object defining the vector's storage properties.

        Returns
        -------
        Dict
            Json response as a dict. Successful request returns `{"acknowledged": true}`.
        """
        body = {
            "properties": {
                field: mapping.to_dict()
            }
        }
        return self.es.transport.perform_request("PUT", f"/{index}/_mapping", body=body)

    def index(self, index: str, field: str, vecs: Iterable[Vec.Base], ids: List[str] = None, refresh: bool = False) -> Tuple[int, List[Dict]]:
        """Index (i.e. store) the given vectors at the given index and field with the optional ids.

        Parameters
        ----------
        index : string
            Index where the vectors are stored.
        field : string
            Field containing the vector in each document.
        vecs : List of `Vec.Base`
            Vectors that should be indexed.
        ids : List of strings
            Optional list of ids associated with the given vectors.
            If not given, the ids are generated by Elasticsearch.
        refresh : bool
            Whether to refresh before returning. Set to true if you want to immediately run queries after indexing.

        Returns
        -------
        Int
            Number of vectors successfully indexed.
        List of Dicts
            Error responses for the vectors that were not indexed.
        """
        if ids is None or ids == []:
            ids = [None for _ in vecs]

        def gen():
            for vec, _id in zip(vecs, ids):
                d = { "_op_type": "index", "_index": index, field: vec.to_dict()}
                if _id:
                    d["_id"] = str(_id)
                elif "_id" in d:
                    del d["_id"]
                yield d

        res = bulk(self.es, gen())
        if refresh:
            self.es.indices.refresh(index=index)
        return res

    def nearest_neighbors(self, index: str, query: NearestNeighborsQuery.Base, k: int = 10, fetch_source: bool = False) -> Dict:
        """Build and execute a nearest neighbors query against the given index.

        Parameters
        ----------
        index : string
            Index to run the search against.
        query : NearestNeighborsQuery.Base
            Query object defining the query properties.
        k: int
            Number of hits to return.
        fetch_source : bool
            Whether to return the `_source` of the document. It's generally faster to _not_ return the `_source`.

        Returns
        -------
        Dict
            Standard Elasticsearch search response parsed as a dict.
        """
        body = {
            "query": {
                "elastiknn_nearest_neighbors": query.to_dict()
            }
        }
        return self.es.search(index, body=body, size=k, _source=fetch_source)
